
#define	CONF_AX25IPD_FILE	AX25_SYSCONFDIR"/ax25ipd.conf"

#define	CONF_AX25ROUTED_FILE	AX25_SYSCONFDIR"/ax25rtd.conf"
#define	DATA_AX25ROUTED_CTL_SOCK AX25_LOCALSTATEDIR"/ax25rtd/control"
#define	PROC_AX25_FILE		"/proc/net/ax25"
#define	DATA_AX25ROUTED_AXRT_FILE	AX25_LOCALSTATEDIR"/ax25rtd/ax25_route"
#define	DATA_AX25ROUTED_IPRT_FILE	AX25_LOCALSTATEDIR"/ax25rtd/ip_route"

#define	PROC_IP_ROUTE_FILE	"/proc/net/route"

#define	PROC_NR_NODES_FILE	"/proc/net/nr_nodes"

#define	CONF_NODE_MOTD_FILE	AX25_SYSCONFDIR"/node.motd"
#define	CONF_NODE_PERMS_FILE	AX25_SYSCONFDIR"/node.perms"
#define	CONF_NODE_FILE		AX25_SYSCONFDIR"/node.conf"

#define	DATA_MHEARD_FILE	AX25_LOCALSTATEDIR"/mheard/mheard.dat"
#define	DATA_NODE_HELP_DIR	"/usr/lib/ax25/node/help/"
#define	CONF_NODE_INFO_FILE	AX25_SYSCONFDIR"/node.info"

#define	DATA_NODE_LOGIN_FILE	AX25_LOCALSTATEDIR"/node/loggedin"

#define	PROC_AX25_CALLS_FILE	"/proc/net/ax25_calls"

#define	MY_TALK			"/usr/sbin/ttylinkd"

#define	MAIL_DELIVERY_AGENT	"/usr/sbin/sendmail %s"

#define	DATA_PMS_LOGIN_FILE	AX25_LOCALSTATEDIR"/pms/loggedin"

#define	PROC_AX25_ROUTE_FILE	"/proc/net/ax25_route"

#define	CONF_PMS_MOTD_FILE	AX25_SYSCONFDIR"/pms.motd"

#define	CONF_PMS_INFO_FILE	AX25_SYSCONFDIR"/pms.info"
